/* _NVRM_COPYRIGHT_BEGIN_
 *
 * Copyright 2014-2016 by NVIDIA Corporation.  All rights reserved.  All
 * information contained herein is proprietary and confidential to NVIDIA
 * Corporation.  Any use, reproduction, or disclosure without the written
 * permission of NVIDIA Corporation is prohibited.
 *
 * _NVRM_COPYRIGHT_END_
 */

#include <linux/module.h>
#include <linux/interrupt.h>
#include "conftest.h"

#include "nvlink_pci.h"
#include "nvlink_proto.h"
#include "nvlink_common.h"
#include "nvlink_errors.h"

#include "ibmnpu_export.h"

#if defined(NVCPU_PPC64LE) && defined(NV_LINUX_OF_H_PRESENT) && \
  defined(NV_OF_PARSE_PHANDLE_PRESENT)
#include <linux/of.h>

#ifndef IRQ_RETVAL
typedef void irqreturn_t;
#define IRQ_RETVAL(a)
#endif

#if !defined(IRQF_SHARED)
#define IRQF_SHARED SA_SHIRQ
#endif

static int              ibmnpu_probe    (struct pci_dev *, const struct pci_device_id *);
static void             ibmnpu_remove   (struct pci_dev *);
static pci_ers_result_t ibmnpu_pci_error_detected   (struct pci_dev *, enum pci_channel_state);
static pci_ers_result_t ibmnpu_pci_mmio_enabled     (struct pci_dev *);

static struct pci_error_handlers ibmnpu_pci_error_handlers =
{
    .error_detected = ibmnpu_pci_error_detected,
    .mmio_enabled   = ibmnpu_pci_mmio_enabled,
};

static struct pci_device_id ibmnpu_pci_table[] = {
    {
        .vendor      = PCI_VENDOR_ID_IBM,
        .device      = PCI_DEVICE_ID_IBM_NPU,
        .subvendor   = PCI_ANY_ID,
        .subdevice   = PCI_ANY_ID,
        .class       = (PCI_CLASS_BRIDGE_NPU << 8),
        .class_mask  = ~1
    },
    { }
};

static struct pci_driver ibmnpu_pci_driver = {
    .name           = IBMNPU_DRIVER_NAME,
    .id_table       = ibmnpu_pci_table,
    .probe          = ibmnpu_probe,
    .remove         = ibmnpu_remove,
    .err_handler    = &ibmnpu_pci_error_handlers,
};




static irqreturn_t ibmnpu_isr
(
    int   irq,
    void *arg
#if !defined(NV_IRQ_HANDLER_T_PRESENT) || (NV_IRQ_HANDLER_T_ARGUMENT_COUNT == 3)
    ,struct pt_regs *regs
#endif
)
{
    NvBool rm_handled = NV_FALSE;

    // TODO: Route the interrupt to NPU Core.

    return IRQ_RETVAL(rm_handled);
}

static int ibmnpu_probe
(
    struct pci_dev *dev,
    const struct pci_device_id *id_table
)
{
    NvlStatus retval = NVL_SUCCESS;
    struct device_node *pci_dev_node;

    NVLINK_PRINT((DBG_MODULE_IBMNPU, NVL_DBG_SETUP, "IBMNPU: Probing Emulated device %04x:%02x:%02x.%x, "
              "Vendor Id = 0x%x, Device Id = 0x%x, Class = 0x%x \n",
              NV_PCI_DOMAIN_NUMBER(dev), NV_PCI_BUS_NUMBER(dev),
              NV_PCI_SLOT_NUMBER(dev), PCI_FUNC(dev->devfn),
              dev->vendor, dev->device, dev->class));

    // Try to register the device in nvlink core library
    retval = ibmnpu_lib_register_device(NV_PCI_DOMAIN_NUMBER(dev),
                                        NV_PCI_BUS_NUMBER(dev),
                                        NV_PCI_SLOT_NUMBER(dev),
                                        PCI_FUNC(dev->devfn),
                                        dev);

    // If there is no GPU associated with this NPU, skip it.
    if (NVL_UNBOUND_DEVICE == retval)
    {
        NVLINK_PRINT((DBG_MODULE_IBMNPU, NVL_DBG_SETUP, "IBMNPU: No GPU is associated to this brick, skipping.\n"));
        return -ENODEV;
    }

    if (NVL_SUCCESS != retval)
    {
        NVLINK_PRINT((DBG_MODULE_IBMNPU, NVL_DBG_ERRORS,
                    "Failed to register NPU device : %d\n",
                    retval));

        goto register_device_fail;
    }

    if (dev->irq == 0)
    {
        NVLINK_PRINT((DBG_MODULE_IBMNPU, NVL_DBG_ERRORS, "IBMNPU: Can't find an IRQ!\n"));
    }

    return 0;

register_device_fail:

    return -1;
}

void ibmnpu_remove(struct pci_dev *dev)
{
    // TODO not supported yet

    NVLINK_PRINT((DBG_MODULE_IBMNPU, NVL_DBG_SETUP, "IBMNPU: removing device %04x:%02x:%02x.%x\n",
              NV_PCI_DOMAIN_NUMBER(dev), NV_PCI_BUS_NUMBER(dev),
              NV_PCI_SLOT_NUMBER(dev), PCI_FUNC(dev->devfn)));
    return;
}

static pci_ers_result_t ibmnpu_pci_error_detected
(
    struct pci_dev *dev,
    enum pci_channel_state error
)
{
    if (NULL == dev)
    {
        return PCI_ERS_RESULT_NONE;
    }

    NVLINK_PRINT((DBG_MODULE_IBMNPU, NVL_DBG_ERRORS, "IBMNPU: ibmnpu_pci_error_detected device %04x:%02x:%02x.%x\n",
                 NV_PCI_DOMAIN_NUMBER(dev), NV_PCI_BUS_NUMBER(dev),
                 NV_PCI_SLOT_NUMBER(dev), PCI_FUNC(dev->devfn)));

    if (pci_channel_io_perm_failure == error)
    {
        return PCI_ERS_RESULT_DISCONNECT;
    }

    //
    // For NPU devices we need to determine if its FREEZE/FENCE EEH, which
    // requires a register read.
    // Tell Linux to continue recovery of the device. The kernel will enable
    // MMIO for the NPU and call the mmio_enabled callback.
    //
    return PCI_ERS_RESULT_CAN_RECOVER;
}

static pci_ers_result_t ibmnpu_pci_mmio_enabled
(
    struct pci_dev *dev
)
{
    if (NULL == dev)
    {
        return PCI_ERS_RESULT_NONE;
    }

    NVLINK_PRINT((DBG_MODULE_IBMNPU, NVL_DBG_ERRORS, "IBMNPU: ibmnpu_pci_mmio_enabled device %04x:%02x:%02x.%x\n",
                 NV_PCI_DOMAIN_NUMBER(dev), NV_PCI_BUS_NUMBER(dev),
                 NV_PCI_SLOT_NUMBER(dev), PCI_FUNC(dev->devfn)));

    //
    // It is understood that we will not attempt to recover from an EEH, but 
    // IBM has requested that we indicate in the logs that it occured and
    // that it was either a FREEZE or a FENCE.
    //
    // Within the MMIO handler specifically, a persistent failure condition
    // is considered a FENCE condition which requires a system power cycle.
    //
    if (ibmnpu_lib_check_failure(dev))
    {
        NVLINK_PRINT((DBG_MODULE_IBMNPU, NVL_DBG_ERRORS, "IBMNPU: NPU FENCE detected, machine power cycle required.\n"));
    }
    else
    {
        NVLINK_PRINT((DBG_MODULE_IBMNPU, NVL_DBG_ERRORS, "IBMNPU: NPU FREEZE detected, driver reload required.\n"));
    }

    NVLINK_PRINT((DBG_MODULE_IBMNPU, NVL_DBG_ERRORS, "IBMNPU: Disconnecting device %04x:%02x:%02x.%x\n",
                 NV_PCI_DOMAIN_NUMBER(dev), NV_PCI_BUS_NUMBER(dev),
                 NV_PCI_SLOT_NUMBER(dev), PCI_FUNC(dev->devfn)));

    // There is no way out at this point, request a disconnect.
    return PCI_ERS_RESULT_DISCONNECT;
}

NvBool NVLINK_API_CALL ibmnpu_lib_check_failure(void *handle)
{
    NvU16 pci_vendor; 

    //
    // According to IBM, any config cycle read of all Fs will cause the
    // firmware to check for an EEH failure on the associated device.
    // If the EEH failure condition exists, EEH error handling will be
    // triggered and PCIBIOS_DEVICE_NOT_FOUND will be returned.    
    //
    return (pci_read_config_word(handle, PCI_VENDOR_ID, &pci_vendor) == PCIBIOS_DEVICE_NOT_FOUND) ? NV_TRUE : NV_FALSE;
}

int ibmnpu_init(void)
{
    NvlStatus retval = ibmnpu_lib_load(0xFFFFFFFF, 0xFFFFFFFF);

    if (NVL_SUCCESS != retval)
    {
        NVLINK_PRINT((DBG_MODULE_IBMNPU, NVL_DBG_ERRORS,
                    "Failed to load ibmnpu library : %d\n",
                    retval));
        return -1;
    }

    return 0;
}

void ibmnpu_exit(void)
{
    NvlStatus retval = ibmnpu_lib_unload();

    if (NVL_SUCCESS != retval)
    {
        NVLINK_PRINT((DBG_MODULE_IBMNPU, NVL_DBG_ERRORS,
                    "Error occurred while unloading ibmnpu library : %d\n",
                    retval));
    }
}



NvlStatus NVLINK_API_CALL ibmnpu_lib_load
(
    NvU32 accepted_domain,
    NvU32 accepted_link_mask
)
{
    NvlStatus retval = NVL_SUCCESS;
    int rc;

    retval = ibmnpu_lib_initialize(accepted_domain, accepted_link_mask);
    if (NVL_SUCCESS != retval)
    {
        NVLINK_PRINT((DBG_MODULE_IBMNPU, NVL_DBG_ERRORS,
                    "Failed to initialize ibmnpu driver : %d\n",
                    retval));

        goto ibmnpu_lib_initialize_fail;
    }

    rc = pci_register_driver(&ibmnpu_pci_driver);
    if (rc < 0)
    {
        NVLINK_PRINT((DBG_MODULE_IBMNPU, NVL_DBG_ERRORS, 
                    "Failed to register ibmnpu driver : %d\n",
                    rc));

        retval = (NvlStatus)rc;
        goto pci_register_driver_fail;
    }

    return retval;

pci_register_driver_fail:

    ibmnpu_lib_shutdown();
ibmnpu_lib_initialize_fail:

    return retval;
}

NvlStatus NVLINK_API_CALL ibmnpu_lib_initialize_device_pci_bus(void *handle)
{
    NvlStatus retval = NVL_SUCCESS;
    int rc;

    if (NULL == handle)
    {
        return -NVL_BAD_ARGS;
    }

    rc = pci_enable_device(handle);
    if (0 != rc)
    {
        NVLINK_PRINT((DBG_MODULE_IBMNPU, NVL_DBG_ERRORS,
                    "Failed to enable pci device : %d\n",
                    rc));

        return (NvlStatus)rc;
    }

    // Enable bus mastering on the device
    pci_set_master(handle);

    return retval;
}

NvlStatus NVLINK_API_CALL ibmnpu_lib_initialize_device_bar_info(void *handle, nvlink_pci_info *info)
{
    NvlStatus retval = NVL_SUCCESS;
    struct pci_dev *dev = handle;
    unsigned int i, j;
    int rc;

    if (NULL == handle || NULL == info)
    {
        return -NVL_BAD_ARGS;
    }

    if (NULL != info->bars[0].pBar)
    {
        NVLINK_PRINT((DBG_MODULE_IBMNPU, NVL_DBG_WARNINGS,
                    "Cannot map ibmnpu device registers : already initialized.\n"));
        return retval;
    }

    rc = pci_request_regions(dev, IBMNPU_DRIVER_NAME);
    if (rc != 0)
    {
        NVLINK_PRINT((DBG_MODULE_IBMNPU, NVL_DBG_ERRORS,
                    "Failed to request memory regions : %d\n",
                    rc));

        return (NvlStatus)rc;
    }

    for (i = 0, j = 0; i < NVRM_PCICFG_NUM_BARS && j < IBMNPU_MAX_BARS; i++)
    {
        if ((NV_PCI_RESOURCE_VALID(dev, i)) &&
            (NV_PCI_RESOURCE_FLAGS(dev, i) & PCI_BASE_ADDRESS_SPACE)
                == PCI_BASE_ADDRESS_SPACE_MEMORY)
        {
            NvU32 bar = 0;

            info->bars[j].offset = NVRM_PCICFG_BAR_OFFSET(i);
            pci_read_config_dword(dev, info->bars[j].offset, &bar);
            info->bars[j].busAddress = (bar & PCI_BASE_ADDRESS_MEM_MASK);
            if (NV_PCI_RESOURCE_FLAGS(dev, i) & PCI_BASE_ADDRESS_MEM_TYPE_64)
            {
                pci_read_config_dword(dev, info->bars[j].offset + 4, &bar);
                info->bars[j].busAddress |= (((NvU64)bar) << 32);
            }


            info->bars[j].baseAddr = NV_PCI_RESOURCE_START(dev, i);
            info->bars[j].barSize = NV_PCI_RESOURCE_SIZE(dev, i);

            NVLINK_PRINT((DBG_MODULE_IBMNPU, NVL_DBG_INFO,
                        "IBMNPU: Bar%d @ 0x%llx [size=%dK].\n",
                        j,
                        info->bars[j].baseAddr,
                        (info->bars[j].barSize >> 10)));

            // Map registers to kernel address space.
            info->bars[j].pBar = pci_iomap(dev, i, 0);
            if (NULL == info->bars[j].pBar)
            {
                NVLINK_PRINT((DBG_MODULE_IBMNPU, NVL_DBG_ERRORS,
                          "IBMNPU: Unable to map BAR%d registers\n", j));
                
                retval = -NVL_PCI_ERROR;
                goto pci_iomap_fail;
            }
            j++;
        }
    }

    return retval;

pci_iomap_fail:
    // Unmap any bar regions which were successfully mapped.
    ibmnpu_lib_shutdown_device_bar_info(dev, info);
    pci_release_regions(dev);

    return retval;
}

NvlStatus NVLINK_API_CALL ibmnpu_lib_unload(void)
{
    NvlStatus retval = ibmnpu_lib_shutdown();

    if (NVL_SUCCESS != retval)
    {
        NVLINK_PRINT((DBG_MODULE_IBMNPU, NVL_DBG_ERRORS,
                    "Failed to shutdown ibmnpu driver : %d\n",
                    retval));
    }

    pci_unregister_driver(&ibmnpu_pci_driver);

    return retval;
}

NvlStatus NVLINK_API_CALL ibmnpu_lib_initialize_device_interrupt(void *handle, nvlink_pci_info *info)
{
    return NVL_SUCCESS;
}

NvlStatus NVLINK_API_CALL ibmnpu_lib_shutdown_device_interrupt(void *handle, nvlink_pci_info *info)
{
    return NVL_SUCCESS;
}

NvlStatus NVLINK_API_CALL ibmnpu_lib_shutdown_device_bar_info(void *handle, nvlink_pci_info *info)
{
    NvlStatus retval = NVL_SUCCESS;
    unsigned int bar;

    if (NULL == handle || NULL == info)
    {
        return -NVL_BAD_ARGS;
    }

    if (NULL == info->bars[0].pBar)
    {
        NVLINK_PRINT((DBG_MODULE_IBMNPU, NVL_DBG_WARNINGS,
                    "Cannot unmap ibmnpu device bars: not initialized.\n"));
        return retval;
    }

    pci_release_regions(handle);

    for (bar = 0; bar < IBMNPU_MAX_BARS; bar++)
    {
        if (NULL != info->bars[bar].pBar)
        {
            pci_iounmap(handle, info->bars[bar].pBar);
            info->bars[bar].pBar = NULL;
        }
    }

    return retval;
}

NvlStatus NVLINK_API_CALL ibmnpu_lib_release_device(void *handle)
{
    if (NULL == handle)
    {
        return -NVL_BAD_ARGS;
    }

    // Placing this here because there is no arch layer "shutdown pci bus"
    pci_disable_device(handle);

    return NVL_SUCCESS;
}

NvU8 NVLINK_API_CALL ibmnpu_lib_pci_read_08 (void *handle, NvU32 offset)
{
    NvU8 buffer = 0xFF;
    if (NULL == handle || offset > NV_PCIE_CFG_MAX_OFFSET)
    {
        return buffer;
    }

    pci_read_config_byte(handle, offset, &buffer);
    return buffer;
}

NvU16 NVLINK_API_CALL ibmnpu_lib_pci_read_16 (void *handle, NvU32 offset)
{
    NvU16 buffer = 0xFFFF;
    if (NULL == handle || offset > NV_PCIE_CFG_MAX_OFFSET)
    {
        return buffer;
    }

    pci_read_config_word(handle, offset, &buffer);
    return buffer;
}

NvU32 NVLINK_API_CALL ibmnpu_lib_pci_read_32 (void *handle, NvU32 offset)
{
    NvU32 buffer = 0xFFFFFFFF;
    if (NULL == handle || offset > NV_PCIE_CFG_MAX_OFFSET)
    {
        return buffer;
    }

    pci_read_config_dword(handle, offset, &buffer);
    return buffer;
}

void  NVLINK_API_CALL ibmnpu_lib_pci_write_08(void *handle, NvU32 offset, NvU8  data)
{
    if (NULL == handle || offset > NV_PCIE_CFG_MAX_OFFSET)
    {
        return;
    }
    
    pci_write_config_byte(handle, offset, data);
}

void  NVLINK_API_CALL ibmnpu_lib_pci_write_16(void *handle, NvU32 offset, NvU16 data)
{
    if (NULL == handle || offset > NV_PCIE_CFG_MAX_OFFSET)
    {
        return;
    }
    
    pci_write_config_word(handle, offset, data);
}

void  NVLINK_API_CALL ibmnpu_lib_pci_write_32(void *handle, NvU32 offset, NvU32 data)
{
    if (NULL == handle || offset > NV_PCIE_CFG_MAX_OFFSET)
    {
        return;
    }
    
     pci_write_config_dword(handle, offset, data);
}

#else

int ibmnpu_init(void)
{
    return 0;
}

void ibmnpu_exit(void)
{
    return;
}

NvlStatus NVLINK_API_CALL ibmnpu_lib_load
(
    NvU32 accepted_domain,
    NvU32 accepted_link_mask
)
{
    return NVL_SUCCESS;
}

NvlStatus NVLINK_API_CALL ibmnpu_lib_initialize_device_pci_bus(void *handle)
{
    return NVL_SUCCESS;
}

NvlStatus NVLINK_API_CALL ibmnpu_lib_initialize_device_bar_info(void *handle, nvlink_pci_info *info)
{
    return NVL_SUCCESS;
}

NvlStatus NVLINK_API_CALL ibmnpu_lib_initialize_device_interrupt(void *handle, nvlink_pci_info *info)
{
    return NVL_SUCCESS;
}

NvlStatus NVLINK_API_CALL ibmnpu_lib_shutdown_device_interrupt(void *handle, nvlink_pci_info *info)
{
    return NVL_SUCCESS;
}

NvlStatus NVLINK_API_CALL ibmnpu_lib_unload(void)
{
    return NVL_SUCCESS;
}

NvlStatus NVLINK_API_CALL ibmnpu_lib_shutdown_device_bar_info(void *handle, nvlink_pci_info *info)
{
    return NVL_SUCCESS;
}

NvlStatus NVLINK_API_CALL ibmnpu_lib_release_device(void *handle)
{
    return NVL_SUCCESS;
}

NvU8  NVLINK_API_CALL ibmnpu_lib_pci_read_08 (void *handle, NvU32 offset)
{
    return 0;
}

NvU16 NVLINK_API_CALL ibmnpu_lib_pci_read_16 (void *handle, NvU32 offset)
{
    return 0;
}

NvU32 NVLINK_API_CALL ibmnpu_lib_pci_read_32 (void *handle, NvU32 offset)
{
    return 0;
}

void  NVLINK_API_CALL ibmnpu_lib_pci_write_08(void *handle, NvU32 offset, NvU8  data)
{
    return;
}

void  NVLINK_API_CALL ibmnpu_lib_pci_write_16(void *handle, NvU32 offset, NvU16 data)
{
    return;
}

void  NVLINK_API_CALL ibmnpu_lib_pci_write_32(void *handle, NvU32 offset, NvU32 data)
{
    return;
}

NvBool NVLINK_API_CALL ibmnpu_lib_check_failure(void *handle)
{
    return 0;
}

#endif
